package dev.fyloz.colorrecipesexplorer.model

import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.validation.NullOrNotBlank
import dev.fyloz.colorrecipesexplorer.model.validation.NullOrSize
import org.hibernate.annotations.ColumnDefault
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull
import javax.validation.constraints.Size

private const val MATERIAL_TYPE_ID_NULL_MESSAGE = "Un identifiant est requis"
private const val MATERIAL_TYPE_NAME_NULL_MESSAGE = "Un nom est requis"
private const val MATERIAL_TYPE_PREFIX_NULL_MESSAGE = "Un préfixe est requis"
private const val MATERIAL_TYPE_PREFIX_SIZE_MESSAGE = "Le préfixe doit faire exactement 3 caractères"

@Entity
@Table(name = "material_type")
data class MaterialType(
  @Id
  @GeneratedValue(strategy = GenerationType.IDENTITY)
  override val id: Long? = null,

  @Column(unique = true)
  override val name: String = "",

  @Column(unique = true)
  val prefix: String = "",

  @Column(name = "use_percentages")
  @ColumnDefault("false")
  val usePercentages: Boolean = false,

  @Column(name = "system_type")
  @ColumnDefault("false")
  val systemType: Boolean = false
) : NamedModel

open class MaterialTypeSaveDto(
  @field:NotBlank(message = MATERIAL_TYPE_NAME_NULL_MESSAGE)
  val name: String,

  @field:NotBlank(message = MATERIAL_TYPE_PREFIX_NULL_MESSAGE)
  @field:Size(min = 3, max = 3, message = MATERIAL_TYPE_PREFIX_SIZE_MESSAGE)
  val prefix: String,

  val usePercentages: Boolean = false
) : EntityDto<MaterialType> {
    override fun toEntity(): MaterialType =
      MaterialType(null, name, prefix, usePercentages)
}

open class MaterialTypeUpdateDto(
  @field:NotNull(message = MATERIAL_TYPE_ID_NULL_MESSAGE)
  val id: Long,

  @field:NullOrNotBlank(message = MATERIAL_TYPE_NAME_NULL_MESSAGE)
  val name: String?,

  @field:NullOrSize(min = 3, max = 3, message = MATERIAL_TYPE_PREFIX_NULL_MESSAGE)
  val prefix: String?
) : EntityDto<MaterialType> {
    override fun toEntity(): MaterialType =
      MaterialType(id, name ?: "", prefix ?: "")
}

// ==== DSL ====
fun materialType(
  id: Long? = null,
  name: String = "name",
  prefix: String = "PRE",
  usePercentages: Boolean = false,
  systemType: Boolean = false,
  op: MaterialType.() -> Unit = {}
) = MaterialType(id, name, prefix, usePercentages, systemType).apply(op)

fun materialType(
  materialType: MaterialType,
  newId: Long? = null,
  newName: String? = null,
  newSystemType: Boolean? = null
) = with(materialType) {
    MaterialType(
      newId ?: id,
      newName ?: name,
      prefix,
      usePercentages,
      newSystemType ?: systemType
    )
}

fun materialTypeSaveDto(
  name: String = "name",
  prefix: String = "PRE",
  usePercentages: Boolean = false,
  op: MaterialTypeSaveDto.() -> Unit = {}
) = MaterialTypeSaveDto(name, prefix, usePercentages).apply(op)

fun materialTypeUpdateDto(
  id: Long = 0L,
  name: String? = null,
  prefix: String? = null,
  op: MaterialTypeUpdateDto.() -> Unit = {}
) = MaterialTypeUpdateDto(id, name, prefix).apply(op)

// ==== Exceptions ====
private const val MATERIAL_TYPE_NOT_FOUND_EXCEPTION_TITLE = "Material type not found"
private const val MATERIAL_TYPE_ALREADY_EXISTS_EXCEPTION_TITLE = "Material type already exists"
private const val MATERIAL_TYPE_CANNOT_DELETE_EXCEPTION_TITLE = "Cannot delete material type"
private const val MATERIAL_TYPE_EXCEPTION_ERROR_CODE = "materialtype"

fun materialTypeIdNotFoundException(id: Long) =
  NotFoundException(
    MATERIAL_TYPE_EXCEPTION_ERROR_CODE,
    MATERIAL_TYPE_NOT_FOUND_EXCEPTION_TITLE,
    "A material type with the id $id could not be found",
    id
  )

fun materialTypeNameNotFoundException(name: String) =
  NotFoundException(
    MATERIAL_TYPE_EXCEPTION_ERROR_CODE,
    MATERIAL_TYPE_NOT_FOUND_EXCEPTION_TITLE,
    "A material type with the name $name could not be found",
    name,
    "name"
  )

fun materialTypeIdAlreadyExistsException(id: Long) =
  AlreadyExistsException(
    MATERIAL_TYPE_EXCEPTION_ERROR_CODE,
    MATERIAL_TYPE_ALREADY_EXISTS_EXCEPTION_TITLE,
    "A material type with the id $id already exists",
    id
  )

fun materialTypeNameAlreadyExistsException(name: String) =
  AlreadyExistsException(
    MATERIAL_TYPE_EXCEPTION_ERROR_CODE,
    MATERIAL_TYPE_ALREADY_EXISTS_EXCEPTION_TITLE,
    "A material type with the name $name already exists",
    name,
    "name"
  )

fun materialTypePrefixAlreadyExistsException(prefix: String) =
  AlreadyExistsException(
    MATERIAL_TYPE_EXCEPTION_ERROR_CODE,
    MATERIAL_TYPE_ALREADY_EXISTS_EXCEPTION_TITLE,
    "A material type with the prefix $prefix already exists",
    prefix,
    "prefix"
  )

fun cannotDeleteMaterialTypeException(materialType: MaterialType) =
  CannotDeleteException(
    MATERIAL_TYPE_EXCEPTION_ERROR_CODE,
    MATERIAL_TYPE_CANNOT_DELETE_EXCEPTION_TITLE,
    "Cannot delete material type ${materialType.name} because one or more materials depends on it"
  )
