package dev.fyloz.colorrecipesexplorer.model.account

import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.exception.RestException
import dev.fyloz.colorrecipesexplorer.model.EntityDto
import dev.fyloz.colorrecipesexplorer.model.Model
import dev.fyloz.colorrecipesexplorer.model.NamedModel
import org.hibernate.annotations.Fetch
import org.hibernate.annotations.FetchMode
import org.springframework.http.HttpStatus
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull
import javax.validation.constraints.Size

private const val GROUP_ID_NULL_MESSAGE = "Un identifiant est requis"
private const val GROUP_NAME_NULL_MESSAGE = "Un nom est requis"
private const val GROUP_PERMISSIONS_EMPTY_MESSAGE = "Au moins une permission est requise"

@Entity
@Table(name = "user_group")
data class Group(
  @Id
  @GeneratedValue(strategy = GenerationType.IDENTITY)
  override var id: Long? = null,

  @Column(unique = true)
  override val name: String = "",

  @Enumerated(EnumType.STRING)
  @ElementCollection(fetch = FetchType.EAGER)
  @CollectionTable(name = "group_permission", joinColumns = [JoinColumn(name = "group_id")])
  @Column(name = "permission")
  @Fetch(FetchMode.SUBSELECT)
  val permissions: MutableSet<Permission> = mutableSetOf(),
) : NamedModel {
    val flatPermissions: Set<Permission>
        get() = this.permissions
          .flatMap { it.flat() }
          .filter { !it.deprecated }
          .toSet()
}

open class GroupSaveDto(
  @field:NotBlank(message = GROUP_NAME_NULL_MESSAGE)
  @field:Size(min = 3)
  val name: String,

  @field:Size(min = 1, message = GROUP_PERMISSIONS_EMPTY_MESSAGE)
  val permissions: MutableSet<Permission>
) : EntityDto<Group> {
    override fun toEntity(): Group =
      Group(null, name, permissions)
}

open class GroupUpdateDto(
  @field:NotNull(message = GROUP_ID_NULL_MESSAGE)
  val id: Long,

  @field:NotBlank(message = GROUP_NAME_NULL_MESSAGE)
  @field:Size(min = 3)
  val name: String,

  @field:Size(min = 1, message = GROUP_PERMISSIONS_EMPTY_MESSAGE)
  val permissions: MutableSet<Permission>
) : EntityDto<Group> {
    override fun toEntity(): Group =
      Group(id, name, permissions)
}

data class GroupOutputDto(
  override val id: Long,
  val name: String,
  val permissions: Set<Permission>,
  val explicitPermissions: Set<Permission>
): Model

fun group(
  id: Long? = null,
  name: String = "name",
  permissions: MutableSet<Permission> = mutableSetOf(),
  op: Group.() -> Unit = {}
) = Group(id, name, permissions).apply(op)

fun groupSaveDto(
  name: String = "name",
  permissions: MutableSet<Permission> = mutableSetOf(),
  op: GroupSaveDto.() -> Unit = {}
) = GroupSaveDto(name, permissions).apply(op)

fun groupUpdateDto(
  id: Long = 0L,
  name: String = "name",
  permissions: MutableSet<Permission> = mutableSetOf(),
  op: GroupUpdateDto.() -> Unit = {}
) = GroupUpdateDto(id, name, permissions).apply(op)

// ==== Exceptions ====
private const val GROUP_NOT_FOUND_EXCEPTION_TITLE = "Group not found"
private const val GROUP_ALREADY_EXISTS_EXCEPTION_TITLE = "Group already exists"
private const val GROUP_EXCEPTION_ERROR_CODE = "group"

class NoDefaultGroupException : RestException(
  "nodefaultgroup",
  "No default group",
  HttpStatus.NOT_FOUND,
  "No default group cookie is defined in the current request"
)

fun groupIdNotFoundException(id: Long) =
  NotFoundException(
    GROUP_EXCEPTION_ERROR_CODE,
    GROUP_NOT_FOUND_EXCEPTION_TITLE,
    "A group with the id $id could not be found",
    id
  )

fun groupNameNotFoundException(name: String) =
  NotFoundException(
    GROUP_EXCEPTION_ERROR_CODE,
    GROUP_NOT_FOUND_EXCEPTION_TITLE,
    "A group with the name $name could not be found",
    name,
    "name"
  )

fun groupIdAlreadyExistsException(id: Long) =
  AlreadyExistsException(
    GROUP_EXCEPTION_ERROR_CODE,
    GROUP_ALREADY_EXISTS_EXCEPTION_TITLE,
    "A group with the id $id already exists",
    id,
  )

fun groupNameAlreadyExistsException(name: String) =
  AlreadyExistsException(
    GROUP_EXCEPTION_ERROR_CODE,
    GROUP_ALREADY_EXISTS_EXCEPTION_TITLE,
    "A group with the name $name already exists",
    name,
    "name"
  )
