package dev.fyloz.colorrecipesexplorer.model.account

import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.EntityDto
import dev.fyloz.colorrecipesexplorer.model.Model
import dev.fyloz.colorrecipesexplorer.model.validation.NullOrNotBlank
import org.hibernate.annotations.Fetch
import org.hibernate.annotations.FetchMode
import org.springframework.security.core.GrantedAuthority
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder
import org.springframework.security.crypto.password.PasswordEncoder
import java.time.LocalDateTime
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull
import javax.validation.constraints.Size

private const val USER_ID_NULL_MESSAGE = "Un numéro d'utilisateur est requis"
private const val USER_LAST_NAME_EMPTY_MESSAGE = "Un nom est requis"
private const val USER_FIRST_NAME_EMPTY_MESSAGE = "Un prénom est requis"
private const val USER_PASSWORD_EMPTY_MESSAGE = "Un mot de passe est requis"
private const val USER_PASSWORD_TOO_SHORT_MESSAGE = "Le mot de passe doit contenir au moins 8 caractères"

@Entity
@Table(name = "user")
data class User(
    @Id
    override val id: Long,

    @Column(name = "first_name")
    val firstName: String = "",

    @Column(name = "last_name")
    val lastName: String = "",

    val password: String = "",

    @Column(name = "default_group_user")
    val isDefaultGroupUser: Boolean = false,

    @Column(name = "system_user")
    val isSystemUser: Boolean = false,

    @ManyToOne
    @JoinColumn(name = "group_id")
    @Fetch(FetchMode.SELECT)
    var group: Group? = null,

    @Enumerated(EnumType.STRING)
    @ElementCollection(fetch = FetchType.EAGER)
    @CollectionTable(name = "user_permission", joinColumns = [JoinColumn(name = "user_id")])
    @Column(name = "permission")
    @Fetch(FetchMode.SUBSELECT)
    val permissions: MutableSet<Permission> = mutableSetOf(),

    @Column(name = "last_login_time")
    var lastLoginTime: LocalDateTime? = null
) : Model {
    val flatPermissions: Set<Permission>
        get() = permissions
            .flatMap { it.flat() }
            .filter { !it.deprecated }
            .toMutableSet()
            .apply {
                if (group != null) this.addAll(group!!.flatPermissions)
            }

    val authorities: Set<GrantedAuthority>
        get() = flatPermissions.map { it.toAuthority() }.toMutableSet()
}

/** DTO for creating users. Allows a [password] a [groupId]. */
open class UserSaveDto(
    @field:NotNull(message = USER_ID_NULL_MESSAGE)
    val id: Long,

    @field:NotBlank(message = USER_FIRST_NAME_EMPTY_MESSAGE)
    val firstName: String,

    @field:NotBlank(message = USER_LAST_NAME_EMPTY_MESSAGE)
    val lastName: String,

    @field:NotBlank(message = USER_PASSWORD_EMPTY_MESSAGE)
    @field:Size(min = 8, message = USER_PASSWORD_TOO_SHORT_MESSAGE)
    val password: String,

    val groupId: Long?,

    @Enumerated(EnumType.STRING)
    val permissions: MutableSet<Permission> = mutableSetOf()
) : EntityDto<User>

open class UserUpdateDto(
    @field:NotNull(message = USER_ID_NULL_MESSAGE)
    val id: Long,

    @field:NullOrNotBlank(message = USER_FIRST_NAME_EMPTY_MESSAGE)
    val firstName: String?,

    @field:NullOrNotBlank(message = USER_LAST_NAME_EMPTY_MESSAGE)
    val lastName: String?,

    val groupId: Long?,

    @Enumerated(EnumType.STRING)
    val permissions: Set<Permission>?
) : EntityDto<User>

data class UserOutputDto(
    override val id: Long,
    val firstName: String,
    val lastName: String,
    val group: Group?,
    val permissions: Set<Permission>,
    val explicitPermissions: Set<Permission>,
    val lastLoginTime: LocalDateTime?
) : Model

data class UserLoginRequest(val id: Long, val password: String)

// ==== DSL ====
fun user(
    passwordEncoder: PasswordEncoder = BCryptPasswordEncoder(),
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    password: String = passwordEncoder.encode("password"),
    isDefaultGroupUser: Boolean = false,
    isSystemUser: Boolean = false,
    group: Group? = null,
    permissions: MutableSet<Permission> = mutableSetOf(),
    lastLoginTime: LocalDateTime? = null,
    op: User.() -> Unit = {}
) = User(
    id,
    firstName,
    lastName,
    password,
    isDefaultGroupUser,
    isSystemUser,
    group,
    permissions,
    lastLoginTime
).apply(op)

fun userSaveDto(
    passwordEncoder: PasswordEncoder = BCryptPasswordEncoder(),
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    password: String = passwordEncoder.encode("password"),
    groupId: Long? = null,
    permissions: MutableSet<Permission> = mutableSetOf(),
    op: UserSaveDto.() -> Unit = {}
) = UserSaveDto(id, firstName, lastName, password, groupId, permissions).apply(op)

fun userUpdateDto(
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    groupId: Long? = null,
    permissions: MutableSet<Permission> = mutableSetOf(),
    op: UserUpdateDto.() -> Unit = {}
) = UserUpdateDto(id, firstName, lastName, groupId, permissions).apply(op)

// ==== Exceptions ====
private const val USER_NOT_FOUND_EXCEPTION_TITLE = "User not found"
private const val USER_ALREADY_EXISTS_EXCEPTION_TITLE = "User already exists"
private const val USER_EXCEPTION_ERROR_CODE = "user"

fun userIdNotFoundException(id: Long) =
    NotFoundException(
        USER_EXCEPTION_ERROR_CODE,
        USER_NOT_FOUND_EXCEPTION_TITLE,
        "An user with the id $id could not be found",
        id
    )

fun userIdAlreadyExistsException(id: Long) =
    AlreadyExistsException(
        USER_EXCEPTION_ERROR_CODE,
        USER_ALREADY_EXISTS_EXCEPTION_TITLE,
        "An user with the id $id already exists",
        id
    )

fun userFullNameAlreadyExistsException(firstName: String, lastName: String) =
    AlreadyExistsException(
        USER_EXCEPTION_ERROR_CODE,
        USER_ALREADY_EXISTS_EXCEPTION_TITLE,
        "An user with the name '$firstName $lastName' already exists",
        "$firstName $lastName",
        "fullName"
    )
