package dev.fyloz.colorrecipesexplorer.utils

/** Returns a list containing the result of the given [transform] applied to each item of the [Iterable]. If the given [transform] throws, the [Throwable] will be passed to the given [throwableConsumer]. */
inline fun <T, R, reified E : Throwable> Iterable<T>.mapMayThrow(
  throwableConsumer: (E) -> Unit = {},
  transform: (T) -> R
): List<R> = this.mapNotNull {
    try {
        transform(it)
    } catch (th: Throwable) {
        if (th is E) {
            throwableConsumer(th)
            null
        } else {
            throw th
        }
    }
}

/** Find duplicated in the given [Iterable] from keys obtained from the given [keySelector]. */
inline fun <T, K> Iterable<T>.findDuplicated(keySelector: (T) -> K) =
  this.groupBy(keySelector)
    .filter { it.value.count() > 1 }
    .map { it.key }

/** Check if the given [Iterable] has gaps between each items, using keys obtained from the given [keySelector]. */
inline fun <T> Iterable<T>.hasGaps(keySelector: (T) -> Int) =
  this.map(keySelector)
    .toIntArray()
    .sorted()
    .filterIndexed { index, it -> it != index + 1 }
    .isNotEmpty()

/** Clears and fills the given [MutableCollection] with the given [elements]. */
fun <T> MutableCollection<T>.setAll(elements: Collection<T>) {
    this.clear()
    this.addAll(elements)
}
